@@ -37,6 +37,7 @@ function Debugger(sel) {
     Handlebars.registerPartial('breakpoint-action-template', $('#breakpoint-action-template').html() );
     Handlebars.registerPartial('breakpoint-right-click-template', $('#breakpoint-right-click-template').html() );
     this.templates = {
+        fileTab: Handlebars.compile( $('#file-tab-template').html() ),
         navTab: Handlebars.compile( $('#nav-tab-template').html() ),
         navTabContent: Handlebars.compile( $('#nav-tab-content-template').html() ),
         navPane: Handlebars.compile( $('#nav-pane-template').html() ),
@@ -47,7 +48,10 @@ function Debugger(sel) {
         breakpointAction: Handlebars.compile( $('#breakpoint-action-template').html() ),
         breakpointListItem: Handlebars.compile( $('#breakpoint-list-item-template').html() ),
         forkedChildModal: Handlebars.compile($('#forked-child-modal-template').html() ),
-        saveLoadBreakpointsModal: Handlebars.compile($('#save-load-breakpoints-modal-template').html() )
+        saveLoadBreakpointsModal: Handlebars.compile($('#save-load-breakpoints-modal-template').html() ),
+        programTerminatedModal: Handlebars.compile($('#program-terminated-modal-template').html() ),
+        subPickerTemplate: Handlebars.compile($('#sub-picker-template').html() ),
+        traceDiffModal: Handlebars.compile($('#trace-diff-modal-template').html() )
     };
 
     // The step in, over, run buttons
@@ -59,7 +63,8 @@ function Debugger(sel) {
                             .addHandler('evalresult', this.watchExpressionResult.bind(this))
                             .addHandler('hangup', this.hangup.bind(this))
                             .addHandler('getvar', this.drawPerlVarPopover.bind(this))
-                            .addHandler('child_process', this.childProcessForked.bind(this));
+                            .addHandler('child_process', this.childProcessForked.bind(this))
+                            .addHandler('trace_diff', this.traceDiff.bind(this));
 
 
         this.getProgramName();
@@ -93,6 +98,10 @@ function Debugger(sel) {
                     this.popoverPerlVar.bind(this));
         this.elt.on('click', '.save-load-breakpoints',
                     this.saveLoadBreakpoints.bind(this));
+        this.elt.on('click', '#add-file',
+                    this.pickFileToLoad.bind(this));
+        this.elt.on('click', '.remove-loaded-file',
+                    this.removeLoadedFile.bind(this));
 
         $('#breakpoint-container-handle').click(function(e) {
             var breakpointPane = $('#breakpoint-container'),
@@ -194,11 +203,30 @@ function Debugger(sel) {
         $(this).css({'height': (($(window).height())-$(this).offset().top)-padding+'px'});
     };
 
+    this.traceDiff = function(data) {
+        var modal = $( this.templates.traceDiffModal(data))
+                        .appendTo(this.elt)
+                        .modal({ backdrop: true, keyboard: true, show: true })
+                        .focus();
+        modal.on('hidden', function(e) {
+            modal.remove();
+        });
+    };
     this.programWasTerminated = function(data) {
-        alert('Debugged program terminated with exit code '+data.exit_code);
-        this.controlButtons.attr('disabled','true');
-        this.controlButtons.filter('#exit').removeAttr('disabled');
+        this.controlButtons.prop('disabled',true);
+        this.controlButtons.find('[data-action="exit"]').prop('disabled', false);
         this.programTerminated = true;
+        var template_data = { };
+        ['exit_code','filename','line','package','exception'].forEach(function(key) {
+            template_data[key] = data[key];
+        });
+        var modal = $( this.templates.programTerminatedModal(template_data))
+                    .appendTo(this.elt)
+                    .modal({ backdrop: true, keyboard: true, show: true})
+                    .focus();
+        modal.on('hidden', function(e) {
+            modal.remove();
+        });
     };
     this.hangup = function(message) {
         var alertMsg = 'Debugged program exited';
@@ -206,7 +234,7 @@ function Debugger(sel) {
             alertMsg = alertMsg + ': ' + message;
         }
         alert(alertMsg);
-        this.controlButtons.attr('disabled','true');
+        this.controlButtons.prop('disabled',true);
         this.programTerminated = true;
     };
 
@@ -508,12 +536,12 @@ function Debugger(sel) {
 
     this.controlButtonClicked = function(e) {
         e.preventDefault();
-        if ($(e.target).attr('disabled')) {
+        if ($(e.target).prop('disabled')) {
             return;
         }
-        dbg.controlButtons.attr('disabled','true');
+        dbg.controlButtons.prop('disabled',true);
 
-        var buttonId = $(e.currentTarget).attr('id');
+        var buttonId = $(e.currentTarget).attr('data-action');
 
         $.ajax({url: buttonId,
                 type: 'GET',
@@ -522,7 +550,7 @@ function Debugger(sel) {
                     dbg.messageHandler.consumer(messages);
                     dbg.refreshAllWatchExpressions();
                     if (! dbg.programTerminated) {
-                        dbg.controlButtons.removeAttr('disabled');
+                        dbg.controlButtons.prop('disabled', false);
                     }
                 },
                 error: this.ajaxError
@@ -539,7 +567,13 @@ function Debugger(sel) {
             eval = $elt.attr('data-eval'),
             stackLevelTag = $elt.closest('.tab-pane').attr('id'),
             stackLevel = 0,
-            dbg = this;
+            dbg = this,
+            timer = $elt.data('timerid');
+
+        if (timer != null) {
+            // mouseout before the timer fired - remove the timer
+            window.clearTimeout(timer);
+        }
 
         // Remove any previous hover var elements
         $('.hovered-perl-var').removeClass('hovered-perl-var');
@@ -563,14 +597,20 @@ function Debugger(sel) {
             $elt.parent().find('[data-eval="'+also+'"]').addClass('hovered-perl-var');
         }
         this.perlVarPopover = $elt;
-        $.ajax({url: 'getvar',
-                dataType: 'json',
-                data: { l: stackLevel, v: eval },
-                type: 'POST',
-                success: this.messageHandler.consumer,
-                error: this.ajaxError
-            });
-    }
+
+        var timer = window.setTimeout(function() {
+            dbg.perlVarPopover.data('timerid', null);
+
+            $.ajax({url: 'getvar',
+                    dataType: 'json',
+                    data: { l: stackLevel, v: eval },
+                    type: 'POST',
+                    success: dbg.messageHandler.consumer,
+                    error: dbg.ajaxError
+                });
+        }, 400);
+        this.perlVarPopover.data('timerid', timer);
+    };
     this.drawPerlVarPopover = function(data) {
         var p,
             popupArgs,
@@ -640,7 +680,7 @@ function Debugger(sel) {
         if (!isVisible) {
             // If the target element is off the top, leave 2 lines at the top,
             // if it's off the bottom, leave 2 lines at the bottom
-            var overShoot = (elemBottom < containerTop) ? -2 : 2;
+            var overShoot = (elemBottom < containerTop) ? 2 : -2;
             container.scrollTo(activeLine, { over: overShoot });
         }
         return activeLine;
@@ -774,6 +814,9 @@ function Debugger(sel) {
         var div = $(e.target).closest('div.watched-expression'),
             expr = div.attr('data-expr');
 
+        if (div.hasClass('editing')) {
+            $('#add-watch-expr').prop('disabled', false);  // re-enable the add button
+        }
         delete this.watchedExpressions[expr];
         div.remove();
     };
@@ -785,29 +828,46 @@ function Debugger(sel) {
         toCollapse.toggle(100);
     };
     this.addEditWatchExpression = function(e) {
-        var newExprInput = $('<input type="text" name="expr" autofocus="1">'),
+        var newExprInput = $('<input type="text" name="expr">'),
             newExprForm = $('<form/>').append(newExprInput),
             watchExpressions = this.elt.find('#watch-expressions-container'),
             $target = $(e.target),
             dbg = this,
             originalExpr,
-            originalElement;
+            originalElement,
+            originalElementHtml;
 
         e.preventDefault();
 
         if ($target.hasClass('expr')) {
             // editing an existing expression
-            originalExpr = $target.html();
             originalElement = $target.closest('div.watched-expression');
+            originalElement.addClass('editing');
+            originalElementHtml = originalElement.html();
+            originalExpr = originalElement.attr('data-expr');
             $target.replaceWith(newExprForm);
             newExprInput.val(originalExpr);
-            newExprForm.__revert = function() { this.replaceWith(originalElement) };
+            newExprForm.__revert = function() {
+                originalElement.html(originalElementHtml);
+                this.__revert = undefined;
+            };
 
         } else {
             //Adding a new expression
             watchExpressions.append(newExprForm);
             newExprForm.__revert = function () { this.remove() };
         }
+        newExprInput.trigger('focus');
+        var addButton = $('#add-watch-expr').prop('disabled', true);  // disable the add button
+
+        newExprInput.keyup(function(e) {
+            // If the user presses escape, abort the form
+            if (e.keyCode == 27) {
+                e.preventDefault();
+                addButton.prop('disabled', false);  // re-enable the add button
+                newExprForm.__revert();
+            }
+        });
 
         newExprForm.submit(function(e) {
             var expr = newExprInput.val();
@@ -815,6 +875,11 @@ function Debugger(sel) {
 
             e.preventDefault();
 
+            addButton.prop('disabled', false);  // re-enable the add button
+            if (originalElement) {
+                originalElement.removeClass('editing');
+            }
+
             if (expr === originalExpr) {
                 newExprForm.__revert();
 
@@ -830,6 +895,7 @@ function Debugger(sel) {
                     // Editing a watch expr
                     delete dbg.watchedExpressions[originalExpr];
                     originalElement.replaceWith(newExpr);
+                    originalElement.attr('data-expr', newExpr);
                 } else {
                     // new watch expr
                     newExprForm.__revert();
@@ -876,6 +942,79 @@ function Debugger(sel) {
         });
     };
 
+    // Pick a tab ID that does not already exist in the #filewindow
+    this.generateNewTabId = function() {
+        var id = 'stack';  // A tab that will always exist to force a pick
+        while( $('#filewindow #file-content #'+id).length) {
+            id = Math.floor(Math.random()*1000000);
+        }
+        return id;
+    }
+
+    // Called when the "+" button in the file tab is clicked
+    this.pickFileToLoad = function(e) {
+        var modal, getSubInfo, loadFile
+            that = this;
+        function closeModal() {
+            modal.remove();
+        }
+        getSubInfo = function(pkg, subname) {
+            modal.modal('hide');
+            var url = 'subinfo/' + pkg + '::' + subname;
+            $.ajax({url: url,
+                    method: 'GET',
+                    dataType: 'json',
+                    success: loadFile
+                });
+        };
+        loadFile = function(subInfo) {
+            var filename = subInfo.data.file,
+                line     = subInfo.data.start;
+
+            that.fileManager.loadFile(filename)
+                            .done( function($elt) { renderNewFile($elt, filename, line) });
+        };
+        renderNewFile = function($codeElt, filename, line) {
+            var tabId = that.generateNewTabId(),
+                newTab = $(that.templates.fileTab({ id: tabId, label: filename, closable: true })),
+                $container = $(that.templates.navPane({tabId: tabId, filename: filename, banner: filename})),
+                $copy = $codeElt.clone().attr('id', tabId);
+
+            // Add the new nav tab
+            $('#add-file').closest('li').before(newTab);
+
+            // add the new tab contents
+            $container.find('.program-code-container').append($copy);
+            $('#filewindow #file-content').append($container);
+            $copy.bind('breakpoints-updated', dbg.breakpointsForPane);
+
+            // Make the new tab active
+            newTab.find('a').click();
+
+            // Set its height
+            dbg._setElementHeight.call($container.find('.managed-height'));
+            // Scroll the requested sub into view
+            dbg._scrollCodeTableLineIntoView($copy, line);
+        };
+        
+        modal = $( this.templates.subPickerTemplate())
+                        .appendTo(this.elt)
+                        .modal({ backdrop: true, keyboard: true, show: true})
+                        .on('hidden', closeModal)
+                        .filePicker({ picked: getSubInfo });
+    };
+
+    this.removeLoadedFile = function(e) {
+        var tabLi = $(e.currentTarget).closest('li'),
+            prevTabLink = tabLi.prev().find('a'),
+            tabId = tabLi.attr('data-contentid'),
+            fileWindow = $('#' + tabId);
+
+        prevTabLink.click();
+        tabLi.remove();
+        fileWindow.remove();
+    };
+
     this.initialize();
 
     function BreakpointManager() {
